#!/usr/bin/env python
# This Python file uses the following encoding: utf-8

# MOPOWER UPS DESIGNED FOR THE RASPBERRY PI 	v0.93 BUILD DATE 11/23/2016 ===BETA===
# Copyright © 2014-2017 All Spectrum Electronics  http://www.allspectrum.com/mopower
# mopower-ups.py for use with the Mopower UPS

# THIS SOFTWARE AND HARDWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# FOR FULL TERMS PLEASE SEE THE LICENSE file included with this software

# To talk to the UPS direct, kill all mopower scripts and run the following...
# minicom -b 19200 -o -D /dev/serial0

#TTY="/dev/ttyAMA0"       # PI 1/2/etc..
#TTY="/dev/ttyS0"         # PI 3
TTY="/dev/serial0"        # THIS SHOULD WORK ON ALL MODELS TO DATE NOV 2016

DETAILLOG="/dev/shm/mopower/mopower-upslog.detail"
#DETAILLOG="/var/log/mopower-upslog.detail"

import time
import sys
import getopt
import serial
import decimal
import datetime
import os

from binascii import hexlify

import crcmod.predefined
crc16 = crcmod.predefined.mkCrcFun('xmodem')

resp = 0;

POWER_COMMAND_TEST = 0

POLYNOMIAL = 0x1021
PRESET = 0

def _initial(c):
    crc = 0
    c = c << 8
    for j in range(8):
        if (crc ^ c) & 0x8000:
            crc = (crc << 1) ^ POLYNOMIAL
        else:
            crc = crc << 1
        c = c << 1
    return crc

_tab = [ _initial(i) for i in range(256) ]

def _update_crc(crc, c):
    cc = 0xff & c

    tmp = (crc >> 8) ^ cc
    crc = (crc << 8) ^ _tab[tmp & 0xff]
    crc = crc & 0xffff
    print (crc)

    return crc

def crc(str):
    crc = PRESET
    for c in str:
        crc = _update_crc(crc, ord(c))
    return crc

def crcb(*i):
    crc = PRESET
    for c in i:
        crc = _update_crc(crc, c)
    return crc

def readlineCR(port):
    rv = ""
    while True:
        ch = port.read()
        rv += ch
        if ch=='\n' or ch=='':
            return rv.strip()

def GET_UPS_VAR(port, PORT_WRITE):
    #port.write("\r\r")
    port.write(PORT_WRITE)
    port.write("\r")
    rv = ""
    while True:
        ch = port.read()
        rv += ch
        #print "=%s=" % rv,
        if ch=='\n' or ch=='':
            #print rv
            RETURN_STRING = rv.split('=', 1 )
            #return rv.strip()
            try:
                DEC_RETURN_STRING = RETURN_STRING[1].strip()
            except IndexError:
                print "============ RETRYING %s ===========" % PORT_WRITE
                DEC_RETURN_STRING = GET_UPS_VAR(port, PORT_WRITE)
                #DEC_RETURN_STRING = 0
            #DEC_RETURN_STRING = float(DEC_RETURN_STRING)
            #return DEC_RETURN_STRING
            return DEC_RETURN_STRING

def GET_UPS_DATA(port, PORT_WRITE):
    #port.write("\r\r")
    port.write(PORT_WRITE)
    port.write("\r")
    rv = ""
    while True:
        ch = port.read()
        rv += ch
        #print "=%s=" % rv,
        if ch=='\n' or ch=='':
            #print rv
            #RETURN_STRING = rv.split('=', 1 )	# THE DIFFERENCE BETWEEN _DATA and _VAR VARIANTS, IS THE _DATA ONE WE DON'T PARSE THE VARIABLE, WE JUST SEND IT AS WE GET IT
            RETURN_STRING = rv
            #return rv.strip()
            try:
                DEC_RETURN_STRING = RETURN_STRING.strip()
            except IndexError:
                print "============ RETRYING %s ===========" % PORT_WRITE
                DEC_RETURN_STRING = GET_UPS_DATA(port, PORT_WRITE)
                #DEC_RETURN_STRING = 0
            #DEC_RETURN_STRING = float(DEC_RETURN_STRING)
            #return DEC_RETURN_STRING
            return DEC_RETURN_STRING

def GET_UPS_DATA_LONG(port, PORT_WRITE):
    #port.write("\r\r")
    port.write(PORT_WRITE)
    port.write("\r")
    rv = ""
    while True:
        ch = port.read(500)
        rv += ch
        #print "=%s=" % rv,
        if ch=='\n' or ch=='':
            #print rv
            #RETURN_STRING = rv.split('=', 1 )  # THE DIFFERENCE BETWEEN _DATA and _VAR VARIANTS, IS THE _DATA ONE WE DON'T PARSE THE VARIABLE, WE JUST SEND IT AS WE GET IT
            RETURN_STRING = rv
            #return rv.strip()
            try:
                DEC_RETURN_STRING = RETURN_STRING.strip()
            except IndexError:
                print "============ RETRYING %s ===========" % PORT_WRITE
                DEC_RETURN_STRING = GET_UPS_DATA(port, PORT_WRITE)
                #DEC_RETURN_STRING = 0
            #DEC_RETURN_STRING = float(DEC_RETURN_STRING)
            #return DEC_RETURN_STRING
            return DEC_RETURN_STRING


def GET_UPS_MULTIPLE_VAR(port, PORT_WRITE):
    #port.write("\r\r")
    port.write(PORT_WRITE)
    port.write("\r")
    rv = ""
    while True:
        ch = port.read()
        rv += ch
        #print "=%s=" % rv,
        if ch=='\n' or ch=='':
            #print rv
            RETURN_STRING = rv
            DEC_RETURN_STRING = RETURN_STRING.strip()

            words = DEC_RETURN_STRING.split()
            #print "=",
            #print words[-1],
            #print "="

            #HEXRESULT = hexlify(words[-1])
            #results = int(HEXRESULT,16)
            #results = words[-1]

            #CRCVALUE = words[-1:]
            try:
                CRCVALUE = int(words[-1], 16)
            except IndexError:
                #print "somethings wrong............................................................."
                CRCVALUE = 0

            VARS_ONLY = DEC_RETURN_STRING[:-5]

            #print "=",
            #print VARS_ONLY,
            #print "="

            #return DEC_RETURN_STRING
            #return VARS_ONLY, words[-1:]
            return VARS_ONLY, CRCVALUE


#port = serial.Serial(TTY, baudrate=19200, timeout=0.1)
port = serial.Serial(TTY, baudrate=19200, timeout=1.0)
port = serial.Serial(TTY, baudrate=19200, timeout=0.1)

def set_time():
    """
    HOUR = time.strftime("%H")
    MIN = time.strftime("%M")
    SEC = time.strftime("%S")

    DAY = time.strftime("%d")
    MONTH = time.strftime("%m")
    YEAR = int(time.strftime("%Y"))
    """
    
    current_time = time.localtime()

    GET_UPS_VAR(port, "HOUR=%s" % current_time.tm_hour)
    GET_UPS_VAR(port, "MIN=%s" % current_time.tm_min)
    GET_UPS_VAR(port, "SEC=%s" % current_time.tm_sec)

    GET_UPS_VAR(port, "DATE=%s" % current_time.tm_mday)
    GET_UPS_VAR(port, "MONTH=%s" % current_time.tm_mon)
    GET_UPS_VAR(port, "YEAR=%s" % current_time.tm_year)

    if ( current_time.tm_wday == 6 ):
        NEWUPS_DAYOFWEEK=1
    else:
        NEWUPS_DAYOFWEEK=current_time.tm_wday + 2
    GET_UPS_VAR(port, "DAYOFWEEK=%s" % NEWUPS_DAYOFWEEK)

def get_cpu_temp():
    tempFile = open( "/sys/class/thermal/thermal_zone0/temp" )
    cpu_temp = tempFile.read()
    tempFile.close()
    return float(cpu_temp)/1000
    # Uncomment the next line if you want the temp in Fahrenheit
    #return float(1.8*cpu_temp)+32

def power_handling():
    global POWER_COMMAND, POWER_COMMAND_TEST
    ##########
    # REBOOT #
    ##########
    if POWER_COMMAND == 2:
        sys.stdout.flush()
	######################SPI_W(30, 0)	# RESET POWER COMMAND BACK TO 0
        import os
	if POWER_COMMAND_TEST == 1:
            print "Reboot command received from MoPower UPS, TEST MODE ONLY, NOT REBOOTING" 
	else:
            print "Reboot command received from MoPower UPS, Rebooting..."
            GET_UPS_VAR(port, "POWER_COMMAND=0")
            GET_UPS_VAR(port, "POWER_COMMAND=0")
	    os.system("sudo reboot")
    #############
    # SHUT DOWN #
    #############
    if POWER_COMMAND == 3:
        sys.stdout.flush()
        #############################SPI_W(30, 0)       # RESET POWER COMMAND BACK TO 0
        time.sleep(1.0)
        import os
        if POWER_COMMAND_TEST == 1:
            print "Shutdown command received from MoPower UPS, TEST MODE ONLY, NOT SHUTTING DOWN"
        else:
            print "Shutdown command received from MoPower UPS, Shutting down..."
            #GET_UPS_VAR(port, "TIMED_SHUTDOWN=20")	# SET POWER CUT TIMER # WE DON'T NORMALLY NEED THIS, SHOULD BE DONE IN HALT/SHUTDOWN SCRIPTS IN OS
            os.system("sudo shutdown -h -P -t now 0")
            time.sleep(10.0)	# sleep for a while so not to double-shutdown

def get_all_data():
   i = 0
   while i < 255:
      SPI_CRC_FAILED=0
      ############################################blatest=SPI(i)
      if SPI_CRC_FAILED == 0 :
         print i, blatest
      i = i + 1


def get_ups_data():
    global SPI_CRC_FAILED, UPS_SEC, UPS_MIN, UPS_HOUR, UPS_DAY, UPS_MONTH, UPS_YEAR, TC_SEC_EXPONENT, TC_SEC_FRACTION, TC_SEC, POWER_COMMAND, TIMED_SHUTDOWN, WATCH, UPTIME, UPTIME_D, ADC_VIN, ADC_VBATT, UPS_YEAR, ADC_P2_1_B1, ADC_P2_1_B2, ADC_P2_1, ADC_P2_3, ADC_P2_5, ADC_P2_7, ADC_VIN_V2, UPS_DOW, SETTINGS_TC_SEC, STATUS_PINB, STATUS_PINC, STATUS_PIND, DDRB, DDRC, DDRD, CRCPASS, SHOW_DATA_OUTPUT, COMPUTEDCRC, CRCVALUE

    #print "Getting UPS Data..",		# DEBUG

    CRCPASS=0

    UPS_SEC = 0
    UPS_MIN = 0
    UPS_HOUR = 0
    UPS_DAY = 0
    UPS_MONTH = 0 
    UPS_YEAR = 0
    UPS_DOW = 0

    SPI_CRC_FAILED=0
    TC_SEC_EXPONENT = 0
    TC_SEC_FRACTION = 0
    TC_SEC = 0
    POWER_COMMAND = 0
    TIMED_SHUTDOWN = 0
    WATCH = 0
    UPTIME = 0
    UPTIME_D = 0
    ADC_VIN = 0
    ADC_VBATT = 0
    ADC_P2_1_B1 = 0
    ADC_P2_1_B2 = 0
    ADC_P2_1 = 0
    ADC_P2_3 = 0
    ADC_P2_5 = 0
    ADC_P2_7 = 0
    ADC_VIN_V2 = 0
    SETTINGS_TC_SEC = 0
    STATUS_PINB = 0
    STATUS_PINC = 0
    STATUS_PIND = 0
    DDRB = 0
    DDRC = 0
    DDRD = 0

    #POWER_COMMAND = GET_UPS_VAR(port, "POWER_COMMAND")
    #TIMED_SHUTDOWN = GET_UPS_VAR(port, "TIMED_SHUTDOWN")
    #UPTIME = GET_UPS_VAR(port, "UPTIME_SEC")
    #UPTIME_D = UPTIME / float(86400)
    #ADC_VIN = GET_UPS_VAR(port, "ADC_VIN_VOLTS")
    #ADC_VBATT = GET_UPS_VAR(port, "ADC_VBATT_VOLTS")
    #SETTINGS_TC_SEC = GET_UPS_VAR(port, "SETTINGS_TC_SEC")
   
    SHOW_DATA_OUTPUT, CRCVALUE = GET_UPS_MULTIPLE_VAR(port, "show data")
    #COMPUTEDCRC=hex(crc16(SHOW_DATA_OUTPUT.encode('utf-8')))
    COMPUTEDCRC=crc16(SHOW_DATA_OUTPUT.encode('utf-8'))

    #print "SHOW_DATA_OUTPUT=%s CRCVALUE=%s COMPUTEDCRC=%s" % (SHOW_DATA_OUTPUT, CRCVALUE, COMPUTEDCRC)

    if ( len(SHOW_DATA_OUTPUT) >= 4 ):
        if ( CRCVALUE == COMPUTEDCRC ):
            #print "-------crc values match---------"
            CRCPASS=1
            BLAPRINT_SPLIT = SHOW_DATA_OUTPUT.split()
            UPS_MONTH = int(BLAPRINT_SPLIT[1])
            UPS_DAY = int(BLAPRINT_SPLIT[2])
            UPS_YEAR = int(BLAPRINT_SPLIT[3])
            UPS_HOUR = int(BLAPRINT_SPLIT[4])
            UPS_MIN = int(BLAPRINT_SPLIT[5])
            UPS_SEC = int(BLAPRINT_SPLIT[6])
            UPS_DOW = int(BLAPRINT_SPLIT[7])
            UPTIME = int(BLAPRINT_SPLIT[8])
            UPTIME_D = UPTIME / float(86400)
            MAIN_POWER = int(BLAPRINT_SPLIT[9])
            UPS_POWER = int(BLAPRINT_SPLIT[10])
            POWER_COMMAND = int(BLAPRINT_SPLIT[11])
            TIMED_SHUTDOWN = int(BLAPRINT_SPLIT[12])
            MAIN_POWER_REQUESTED = int(BLAPRINT_SPLIT[13])
            ADC_VIN = float(BLAPRINT_SPLIT[14])
            ADC_VBATT = float(BLAPRINT_SPLIT[15])
            ADC_VINVBATT_VOLTS_DIFF = float(BLAPRINT_SPLIT[16])
            
            
def write_log(write_filename):
    global VERBOSE
    if write_filename == "sys.stdout" :
        logfile = sys.stdout
    else:
        logfile = open(write_filename, 'a')

    print >> logfile, "%s %s" % (time.strftime("%m/%d/%Y"), time.strftime("%H:%M:%S")),
    #print >> logfile, "%.2d/%.2d/%.4d %.2d:%.2d:%.2d UPS_DOW=%d ADC_VIN=%.2f ADC_VBATT=%.2f WD=%d UPTIME=%d,%.2f POWER_COMMAND=%d TIMED_SHUTDOWN=%d CPU_TEMP=%.2f SETTINGS_TC_SEC=%d CRCPASS=%d" % (UPS_MONTH, UPS_DAY, UPS_YEAR, UPS_HOUR, UPS_MIN, UPS_SEC, UPS_DOW, ADC_VIN, ADC_VBATT, WATCH, UPTIME, UPTIME_D, POWER_COMMAND, TIMED_SHUTDOWN, CPU_TEMP, SETTINGS_TC_SEC, CRCPASS),
    print >> logfile, "%.2d/%.2d/%.4d %.2d:%.2d:%.2d VIN=%.2f VBATT=%.2f UPTIME=%d,%.2f PWR_CMD=%d TMD_SHTDWN=%d CPU_TMP=%.2f CRCPASS=%d" % (UPS_MONTH, UPS_DAY, UPS_YEAR, UPS_HOUR, UPS_MIN, UPS_SEC, ADC_VIN, ADC_VBATT, UPTIME, UPTIME_D, POWER_COMMAND, TIMED_SHUTDOWN, CPU_TEMP, CRCPASS),
    if ( VERBOSE == 1 ): 
        print >> logfile, "SHOW_DATA_OUTPUT=%s CRCVALUE=%s COMPUTEDCRC=%s" % (SHOW_DATA_OUTPUT, CRCVALUE, COMPUTEDCRC),
    print >> logfile, ""
    if write_filename != "sys.stdout" :
        logfile.close()

def process_args(argv):
    global sleeptime, VERBOSE, STDOUT
    VERBOSE=0
    STDOUT=1
    sleeptime = 1.0
    outputfile = ''
    try:
        opts, args = getopt.getopt(argv,"hs:o:t:w:z:a:r:vn",["--sleep-time=","ofile="])
    except getopt.GetoptError:
        print 'mopower-log.py -s <sleeptime>'
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
           print 'mopower-log.py -s <sleeptime>'
           sys.exit()
        elif opt in ("-s", "--sleep-time"):
           sleeptime = float(arg)
        elif opt in ("-o", "--ofile"):
           outputfile = arg
        elif opt == '-t':
           print "Setting time..." 
           set_time()
           sys.exit()
        elif opt == '-w':
           print "Resetting watchdog..."
           #####################tempspi = spi.xfer2([253])
           #################### MORE SPI STUFF REMOVED
           sys.exit()
        elif opt == '-z':
           print "Setting shutdown timer..."
           ############################SPI_W(31, 7)  # SET POWER CUT TIMER
           GET_UPS_VAR(port, "TIMED_SHUTDOWN=5")
           sys.exit()
        elif opt == '-a':
           print "Get all settings..."
           ######################################get_all_data()
           sys.exit()
        elif opt == '-r':
           print "Rebooting UPS..."
           port.write("\rUPS-REBOOT\r")
           sys.exit()
        elif opt == '-v':
           VERBOSE=1
        elif opt == '-n':
           STDOUT=0


#port.write("\r")
#port.flushInput()
#port.flushOutput()
#time.sleep(0.1)

port.write("\r")
#time.sleep(0.2)
#throwaway = port.read(9999)

process_args(sys.argv[1:])


if not os.path.exists("/run/shm/mopower/"):
    os.makedirs("/run/shm/mopower/")

FILEHANDLE_TIME_SHUTDOWN = open("/run/shm/mopower/TIMED_SHUTDOWN", 'w')
#print >> logfile, "0"
FILEHANDLE_TIME_SHUTDOWN.close()

FILEHANDLE_COMMAND = open("/run/shm/mopower/COMMAND", 'w')
os.chmod("/run/shm/mopower/COMMAND", 0666)
FILEHANDLE_COMMAND.close()

PREVMIN=100

while True:

    now = datetime.datetime.now()
    #print "Current second: %d" % now.second
    #if (now.second % 2) == 0 :
    if 1 :
        CPU_TEMP = get_cpu_temp()
        
        get_ups_data()

        write_log(DETAILLOG)
        if STDOUT:
            write_log("sys.stdout")
        if PREVMIN != now.minute :
            write_log("/var/log/mopower-upslog") 
            PREVMIN=now.minute
        power_handling() 

    FILE_TIMED_SHUTDOWN = open("/run/shm/mopower/TIMED_SHUTDOWN", 'r')
    DATA_TIMED_SHUTDOWN=FILE_TIMED_SHUTDOWN.read().rstrip()
    if ( DATA_TIMED_SHUTDOWN != "" ) :
        print "UPS COMMAND: TIMED_SHUTDOWN=%s" % DATA_TIMED_SHUTDOWN
        GET_UPS_VAR(port, "TIMED_SHUTDOWN=%s" % DATA_TIMED_SHUTDOWN)
        FILEHANDLE_TIME_SHUTDOWN = open("/run/shm/mopower/TIMED_SHUTDOWN", 'w')
        FILEHANDLE_TIME_SHUTDOWN.close()

    FILEHANDLE_READ_COMMAND = open("/run/shm/mopower/COMMAND", 'r')
    FILEHANDLE_READ_DATA_COMMAND=FILEHANDLE_READ_COMMAND.read().rstrip()
    if ( FILEHANDLE_READ_DATA_COMMAND != "" ) :
        print "RUNNING COMMAND: %s" % FILEHANDLE_READ_DATA_COMMAND
        FILEHANDLE_READ_DATA_COMMAND_OUTPUT = GET_UPS_DATA_LONG(port, "%s" % FILEHANDLE_READ_DATA_COMMAND)
        FILEHANDLE_WRITE_COMMAND_OUTPUT = open("/run/shm/mopower/COMMAND.out", 'w')
        print "%s" % FILEHANDLE_READ_DATA_COMMAND_OUTPUT
        print >> FILEHANDLE_WRITE_COMMAND_OUTPUT, "%s" % FILEHANDLE_READ_DATA_COMMAND_OUTPUT
        FILEHANDLE_WRITE_COMMAND_OUTPUT.close()
        
        FILEHANDLE_COMMAND = open("/run/shm/mopower/COMMAND", 'w')
        FILEHANDLE_COMMAND.close()

    sys.stdout.flush()
    time.sleep(sleeptime)

    # THIS SETS THE UPS TIME WHEN IT DETECTS IT HAS NOT BEEN SET BEFORE, MORE AGRESSIVE WAY TO SET UPS TIME, BUT MAY NOT BE CORRECT, RECOMMEND USE NTP DETECTION BELOW
    #if ( UPS_YEAR == 1970 ) :
	#set_time()

    # THIS SETS THE UPS TIME WHEN WE HAVE CONFIRMATION ON AN NTP TIME SYNC, THIS IS THE PREFERRED METHOD
    if now.second == 0 and (now.minute % 2) == 0 :
        ntpq = os.system("/home/pi/mopower/mopower-ups-ntp-check.sh")
        #print ntpq
        if ntpq == 0:
            print "NTP time sync comfirmed, setting UPS time...\n"
            set_time()

    # Sets system date from UPS if date was not set, and then about every 10 minutes 
    if UPS_YEAR > 1971 :	# CHECK IF UPS_YEAR IS VALID
        if now.year < 1971 or ((now.minute % 10) == 0 and now.second == 0) :
            print "Setting system date from UPS"
            os.system("sudo date -s '%s-%s-%s %s:%s:%s' > /dev/null" % (UPS_YEAR, UPS_MONTH, UPS_DAY, UPS_HOUR, UPS_MIN, UPS_SEC))
